﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using System.Drawing;
using FeatureDetectors;
using System.Windows.Forms;
using Emgu.CV;
using Emgu.CV.Structure;
using Emgu.CV.CvEnum;
using SIFT;
using DotNetMatrix;

namespace FeatureDetectors {
	public class SIFTOptions : CDetectorOptions {
		public int scaleLevels 				= 3;
		public int maxOctaves 				= 5;
		public int minImageSize 			= 32;
		public double curvatureRatio		= 10; 
		public double interpolateThreshold 	= 0.03;
		public double peakThreshold 		= 0.005;
		public bool		doubleImage			= true;
	}
	public class SIFTFeature : Feature {
		public bool processed = false;
	}
	public class OctaveLayer {
		public Image<Gray, float> DoG { get; private set; }
		public Image<Gray, float> Gaus { get; private set; }

		public float[,] Ori { get; private set; }
		public float[,] Grad { get; private set; }

		public int level { get; private set; }
		public List<SIFTFeature> features { get; private set; }
		public double scale { get; private set; }
		public int origHeight;
		public int origWidth;
		public bool filtered = false;

		public OctaveLayer( Image<Gray, float> img, Image<Gray, float> dog, int _level, double _scale ) {
			Gaus = img;
			DoG = dog;
			level = _level;
			scale = _scale;
			features = new List<SIFTFeature>();
		}

		
		public int FindFeatures(List<OctaveLayer> layers, SIFTOptions options) {
			GeneralMatrix offset = new GeneralMatrix(3, 1);
			double peakThreshold = options.peakThreshold *255;
			double interThreshold = options.interpolateThreshold *255;
			double curvatureRatio = Math.Pow(options.curvatureRatio + 1, 2) / options.curvatureRatio;
			byte[,] keyPointMap = new byte[DoG.Height, DoG.Width];
			for ( int x = 1; x < DoG.Width - 1; x++ )
				for ( int y = 1; y < DoG.Height - 1; y++ ) {
					if (Math.Abs(DoG.Data[y, x, 0]) < peakThreshold) continue;
					short maxType = IsExtremalPoint(DoG.Data[y, x, 0], ref x, ref y, true);
					if (maxType == 0) continue;
					if ( maxType != layers[level - 1].IsExtremalPoint(DoG.Data[y, x, 0], ref x, ref y, false) ) continue;
					if ( maxType != layers[level + 1].IsExtremalPoint(DoG.Data[y, x, 0], ref x, ref y, false) ) continue;
					
					if ( FilterHessian(ref x, ref y) > curvatureRatio) continue;

					int iter = 0;
					int nx = x, ny = y;
					double peakVal = InterpolateFeature(ref x, ref y, layers, out offset);
					
					while (offset!=null && iter++<5) {
						if (offset[1, 0]*offset[1, 0] + offset[2, 0]*offset[2, 0]  + offset[0, 0]*offset[0, 0] > 3) {
							iter = 10;
							break;
						}
						if (offset[1,0] < -0.5) ny--;
 						if (offset[1,0] > 0.5) ny++;
 						if (offset[2,0] < -0.5) nx--;
 						if (offset[2,0] > 0.5) nx++;
						if (nx==x && ny == y) break;
						if (nx<1 || ny<1 || nx>=DoG.Width-1 || ny>=DoG.Height-1) {
							iter = 10;
							break;
						}
						
						peakVal = InterpolateFeature(ref nx, ref ny, layers, out offset);
					}
					if (iter>=5 || Math.Abs(peakVal) < interThreshold) continue;
					
					if (keyPointMap[ny, nx] != 0) continue;
					keyPointMap[ny, nx] = 1;
				
					AddFeature(ref nx, ref ny, offset, options);
				}
			return features.Count;
		}

		public double FilterHessian(ref int x, ref int y) {
			double Dxx = DoG.Data[y, x + 1, 0] - 2.0 * DoG.Data[y, x, 0] + DoG.Data[y, x - 1, 0];
			double Dyy = DoG.Data[y + 1, x, 0] - 2.0 * DoG.Data[y, x, 0] + DoG.Data[y - 1, x, 0];
			double Dxy = ((DoG.Data[y + 1, x + 1, 0] - DoG.Data[y + 1, x - 1, 0]) -
				(DoG.Data[y - 1, x + 1, 0] - DoG.Data[y - 1, x - 1, 0])) / 4.0;
			
			// Trace(Hessian)^2 / Det(Hessian) < (r+1)^2/r
			return (Math.Pow(Dxx + Dyy, 2) / (Dxx * Dyy - Dxy * Dxy));
		}
		public double InterpolateFeature(ref int x, ref int y,  List<OctaveLayer> layers, out GeneralMatrix offset) {
		    float[,,] dogUp	 	= layers[level - 1].DoG.Data;
			float[,,] dog	 	= DoG.Data;
			float[,,] dogDown	= layers[level + 1].DoG.Data;
			
			GeneralMatrix g = new GeneralMatrix(3, 1);
			g[0, 0] = (dogDown[y, x, 0] - dogUp[y, x, 0]) / 2.0;
			g[1, 0] = (dog[y + 1, x, 0] - dog[y - 1, x, 0]) / 2.0;
			g[2, 0] = (dog[y, x + 1, 0] - dog[y, x - 1, 0]) / 2.0;
			

			GeneralMatrix H = new GeneralMatrix(3,3);
			H[0, 0] = dogUp[y, x, 0] - 2 * dog[y, x, 0] + dogDown[y, x, 0];
			H[0, 1] = (dogDown[y + 1, x, 0] - dogDown[y - 1, x, 0] - (dogUp[y + 1, x, 0] - dogUp[y - 1, x, 0])) / 4.0;
			H[0, 2] = (dogDown[y, x + 1, 0] - dogDown[y, x - 1, 0] - (dogUp[y, x + 1, 0] - dogUp[y, x - 1, 0])) / 4.0;
			
			H[1, 1] = dog[y - 1, x, 0] - 2 * dog[y, x, 0] + dog[y + 1, x, 0];
			H[1, 2] = (dog[y + 1, x + 1, 0] - dog[y + 1, x - 1, 0] - (dog[y - 1, x + 1, 0] - dog[y - 1, x - 1, 0])) / 4.0;
			
			H[2, 2] = dog[y, x - 1, 0] - 2 * dog[y, x, 0] + dog[y, x + 1, 0];
			
			// symetricity
			H[1, 0] = H[0, 1];
			H[2, 0] = H[0, 2];
 			H[2, 1] = H[1, 2];

			// linear system Hx = -g
			try {
				offset = H.Solve(g.Multiply( -1));
			}
			catch {
				offset = null;
				return 0;
			}
			return dog[y, x, 0] + 0.5*(offset.Transpose()*g)[0, 0];
		}

		public short IsExtremalPoint(float pointValue, ref int X, ref int Y, bool sameLayer = false) {
			float max = -255, min = 255;
			for (int x = X - 1; x <= X + 1; x++ ) {
				for (int y = Y - 1; y <= Y + 1; y++ ) {
					if (sameLayer && (y == Y) && (x == X)) continue;
					max = Math.Max(DoG.Data[y, x, 0], max);
					min = Math.Min(DoG.Data[y, x, 0], min);
					if ( pointValue < max && pointValue > min ) return  0;
				}
			}
			
			return (short) (pointValue > max ? 1 : -1);
		}


		public SIFTFeature AddFeature(ref int x, ref int y, GeneralMatrix offset, SIFTOptions options) {
			SIFTFeature feature = new SIFTFeature();

			feature.position = new Point(x, y);
			// TODO add offset position to location;
			feature.location = new PointF(origWidth*((float) offset.Array[2][0]+x)/DoG.Width, origHeight*((float) offset.Array[1][0]+y)/DoG.Height);

			
			feature.processed = true;
			feature.scale = scale;

			//double deltaX = DoG.Data[y, x + 1, 0] - DoG.Data[y, x - 1, 0];
			//double deltaY = DoG.Data[y + 1, x, 0] - DoG.Data[y - 1, x, 0];
			//m(x, y) = sqrt(L(x + 1, y) − L(x − 1, y))^2 + (L(x, y + 1) − L(x, y − 1))^2
			//feature.magnitude =  Math.Sqrt(Math.Pow(deltaX, 2) + Math.Pow(deltaY, 2));

			//theta(x, y) = tan^−1((L(x, y + 1) − L(x, y − 1))/(L(x + 1, y) − L(x − 1, y)))
			//feature.direction = Math.Atan2( deltaY, deltaX );
			//if (feature.magnitude < options.magnitudeThreshold) return null;
			features.Add(feature);
			return feature;
		}
	}
	public class Octave {
		public List<OctaveLayer> layers { get; private set; }
		public int level { get; private set; }
		public int origWidth;
		public int origHeight;
		public int scaleLevels = 0;

		public Octave() {

		}
		public Octave(Image<Gray, float> image, SIFTOptions options, int width, int height, double imageScale) {
			if ( Math.Min(image.Width, image.Height) < options.minImageSize ) return;
			origWidth = width;
			origHeight = height;
			scaleLevels = options.scaleLevels;
			layers = new List<OctaveLayer>();
			Image<Gray, float> prevImage = image;
			double k = Math.Pow(2, 1 / (double) scaleLevels);
			for ( int i = 0; i < scaleLevels + 2; i++ ) {

				Image<Gray, float> gaus = prevImage.SmoothGaussian(0, 0, imageScale*Math.Sqrt(k*k-1), 0);
				Image<Gray, float> dog = gaus.Sub(prevImage);
				
				OctaveLayer layer = new OctaveLayer(prevImage, dog, layers.Count, imageScale);
				layer.origHeight = origHeight;
				layer.origWidth = origWidth;
				prevImage = gaus;
				layers.Add(layer);
				imageScale *= k;
			}
		}
		public Image<Gray, float> NextImage() {
			Image<Gray, float> img = layers[scaleLevels].Gaus;
			Image<Gray, float> ret = new Image<Gray, float>(img.Width/2, img.Height/2);
			for (int x=0; x<ret.Width; x++)
				for (int y=0; y<ret.Height; y++) ret.Data[y, x, 0] = img.Data[y*2, x*2, 0];
			return ret;
		}
			
		public List<SIFTFeature> GetFeatures(int _scaleLevel) {
			return layers[_scaleLevel].features;
		}

		public List<SIFTFeature> GetFeatures() {
			List<SIFTFeature> res = new List<SIFTFeature>();
			for ( int i=1; i <= scaleLevels; i++ ) res.AddRange(GetFeatures(i));
			return res;
		}

	}
	public class CDetectorFunc : CDetector {
		public override string detectorName { get { return "SIFT"; } }
		private OptionsDialog optDialog;


		public CDetectorFunc() {
			options = new SIFTOptions();
			optDialog = new OptionsDialog((SIFTOptions) options);
		}
		override public DialogResult ShowOptionsDialog() {
			optDialog.Show();
			optDialog.Activate();
			return DialogResult.OK;
		}
		// return 0 is not, 1 if is max, -1 if is min short
		


		override public Feature[] GetFeatures(Image<Gray, byte> image) {
			/*** SIFT by LOWE : Object Recognition from Local Scale-Invariant Features (2003)
			 * 
			 * Patent warning: The University of British Columbia has applied for a patent on the SIFT
			 * algorithm in the United States. Commercial applications of this software may require a 
			 * license from the University of British Columbia. 
			 */

			Image<Gray, float> startImage = image.Convert<Gray, float>();
			
			double imageScale = 0.5;

			int origWidth = image.Width;
			int origHeight = image.Height;

			if (((SIFTOptions) options).doubleImage) {
				startImage = startImage.Resize(startImage.Width * 2, startImage.Height * 2, INTER.CV_INTER_LINEAR);
				imageScale = 1.0;
			}
			List<SIFTFeature> features = new List<SIFTFeature>();
			Octave octave;
			int octaveCount = 0;
			while (octaveCount++ < ((SIFTOptions) options).maxOctaves && (octave = new Octave(startImage, (SIFTOptions) options, origWidth, origHeight, imageScale)).scaleLevels > 0 ) {
				for ( int scaleLevel = 1; scaleLevel <= octave.scaleLevels; scaleLevel++ ) {					
					octave.layers[scaleLevel].FindFeatures(octave.layers, (SIFTOptions) options);
					features.AddRange(octave.layers[scaleLevel].features);
				}
				for (int scaleLevel = 0; scaleLevel <= octave.scaleLevels+1; scaleLevel++) {
					//ShowImage(octave.layers[scaleLevel].Gaus, "GAUS octave: "+octaveCount.ToString() + " scale: "+scaleLevel.ToString());
					//ShowImage(octave.layers[scaleLevel].DoG, "DOG octave: "+octaveCount.ToString() + " scale: "+scaleLevel.ToString());
				}
				startImage = octave.NextImage();
				imageScale *= 2;
			}

			


			return features.ToArray();

		}
	}
}
